@@ -11,12 +11,13 @@
  *****************************************************************************/
 
 
-/*>>>>>>>>>>>>>>>>>>>>> DEFINITIONS & DECLARATIONS <<<<<<<<<<<<<<<<<<<<<<*/
+/********************** DEFINITIONS & DECLARATIONS ***********************/
 
 #include <ipv4.h>
 #include <udp.h>
 #include <tcp.h>
 #include <ethernet.h>
+#include <time.h>
 #include <sys/socket.h>
 #include <string.h>
 
@@ -49,6 +50,7 @@ struct arp_entry {
 	uint8_t  mac_addr[6];
 	uint8_t  eth_frame[ETH_MTU_SIZE];
 	int      eth_len;
+	int	 pkt_pending;
 };
 
 /** \struct icmphdr
@@ -79,32 +81,26 @@ struct icmphdr {
 	} payload;
 };
 
-/*>>>>>>>>>>>>>>>>>>>>>>>>>>>>> PROTOTYPES <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<*/
+/****************************** PROTOTYPES *******************************/
 
-static unsigned short
-checksum(unsigned short *packet, int words);
+static unsigned short checksum(unsigned short *packet, int words);
 
-static void
-arp_send_request(uint32_t dest_ip);
+static void arp_send_request(int fd, uint32_t dest_ip);
 
-static void
-arp_send_reply(uint32_t src_ip, uint8_t * src_mac);
+static void arp_send_reply(int fd, uint32_t src_ip, uint8_t * src_mac);
 
-static void
-fill_arphdr(uint8_t * packet, uint8_t opcode,
-            const uint8_t * src_mac, uint32_t src_ip,
-            const uint8_t * dest_mac, uint32_t dest_ip);
+static void fill_arphdr(uint8_t * packet, uint8_t opcode,
+			const uint8_t * src_mac, uint32_t src_ip,
+			const uint8_t * dest_mac, uint32_t dest_ip);
 
-static arp_entry_t*
-lookup_mac_addr(uint32_t ipv4_addr);
+static arp_entry_t *lookup_mac_addr(uint32_t ipv4_addr);
 
-static void
-fill_udp_checksum(struct iphdr *ipv4_hdr);
+static void fill_udp_checksum(struct iphdr *ipv4_hdr);
 
-static int8_t
-handle_icmp(struct iphdr * iph, uint8_t * packet, int32_t packetsize);
+static int8_t handle_icmp(int fd, struct iphdr * iph, uint8_t * packet,
+			  int32_t packetsize);
 
-/*>>>>>>>>>>>>>>>>>>>>>>>>>>>>> LOCAL VARIABLES <<<<<<<<<<<<<<<<<<<<<<<<<*/
+/****************************** LOCAL VARIABLES **************************/
 
 /* Routing parameters */
 static uint32_t own_ip       = 0;
@@ -125,16 +121,18 @@ static unsigned int arp_consumer = 0;
 static unsigned int arp_producer = 0;
 static arp_entry_t  arp_table[ARP_ENTRIES];
 
-/* Function pointer send_ip. Points either to send_ipv4() or send_ipv6() */
-int   (*send_ip) (void *, int);
+static uint8_t pending_pkt_frame[ETH_MTU_SIZE];
+static int pending_pkt_len;
 
-/*>>>>>>>>>>>>>>>>>>>>>>>>>>>> IMPLEMENTATION <<<<<<<<<<<<<<<<<<<<<<<<<<<*/
+/* Function pointer send_ip. Points either to send_ipv4() or send_ipv6() */
+int   (*send_ip) (int fd, void *, int);
+
+/***************************** IMPLEMENTATION ****************************/
 
 /**
  * IPv4: Initialize the environment for the IPv4 layer.
  */
-static void
-ipv4_init(void)
+static void ipv4_init(void)
 {
 	int i;
 
@@ -147,9 +145,10 @@ ipv4_init(void)
 		arp_table[i].ipv4_addr = 0;
 		memset(arp_table[i].mac_addr, 0, 6);
 		arp_table[i].eth_len = 0;
+		arp_table[i].pkt_pending = 0;
 	}
 
-	/* Set IP send function to send_ipv4() */ 
+	/* Set IP send function to send_ipv4() */
 	send_ip = &send_ipv4;
 }
 
@@ -158,8 +157,7 @@ ipv4_init(void)
  *
  * @param  _own_ip  client IPv4 address (e.g. 127.0.0.1)
  */
-void
-set_ipv4_address(uint32_t _own_ip)
+void set_ipv4_address(uint32_t _own_ip)
 {
 	own_ip = _own_ip;
 	ipv4_init();
@@ -170,8 +168,7 @@ set_ipv4_address(uint32_t _own_ip)
  *
  * @return client IPv4 address (e.g. 127.0.0.1)
  */
-uint32_t
-get_ipv4_address(void)
+uint32_t get_ipv4_address(void)
 {
 	return own_ip;
 }
@@ -181,8 +178,7 @@ get_ipv4_address(void)
  *
  * @param  _own_ip  multicast IPv4 address (224.0.0.0 - 239.255.255.255)
  */
-void
-set_ipv4_multicast(uint32_t _multicast_ip)
+void set_ipv4_multicast(uint32_t _multicast_ip)
 {
 	// is this IP Multicast out of range (224.0.0.0 - 239.255.255.255)
 	if((htonl(_multicast_ip) < 0xE0000000)
@@ -206,8 +202,7 @@ set_ipv4_multicast(uint32_t _multicast_ip)
  *
  * @return multicast IPv4 address (224.0.0.0 - 239.255.255.255 or 0 if not set)
  */
-uint32_t
-get_ipv4_multicast(void)
+uint32_t get_ipv4_multicast(void)
 {
 	return multicast_ip;
 }
@@ -217,8 +212,7 @@ get_ipv4_multicast(void)
  *
  * @param  _router_ip   router IPv4 address
  */
-void
-set_ipv4_router(uint32_t _router_ip)
+void set_ipv4_router(uint32_t _router_ip)
 {
 	router_ip = _router_ip;
 	ipv4_init();
@@ -229,8 +223,7 @@ set_ipv4_router(uint32_t _router_ip)
  *
  * @return router IPv4 address
  */
-uint32_t
-get_ipv4_router(void)
+uint32_t get_ipv4_router(void)
 {
 	return router_ip;
 }
@@ -240,8 +233,7 @@ get_ipv4_router(void)
  *
  * @param  _subnet_mask   netmask of the own IPv4 address
  */
-void
-set_ipv4_netmask(uint32_t _subnet_mask)
+void set_ipv4_netmask(uint32_t _subnet_mask)
 {
 	subnet_mask = _subnet_mask;
 	ipv4_init();
@@ -252,8 +244,7 @@ set_ipv4_netmask(uint32_t _subnet_mask)
  *
  * @return netmask of the own IPv4 address
  */
-uint32_t
-get_ipv4_netmask(void)
+uint32_t get_ipv4_netmask(void)
 {
 	return subnet_mask;
 }
@@ -276,9 +267,9 @@ get_ipv4_netmask(void)
  * @see                fill_dnshdr
  * @see                fill_btphdr
  */
-void
-fill_iphdr(uint8_t * packet, uint16_t packetsize,
-           uint8_t ip_proto, uint32_t ip_src, uint32_t ip_dst) {
+void fill_iphdr(uint8_t * packet, uint16_t packetsize,
+           uint8_t ip_proto, uint32_t ip_src, uint32_t ip_dst)
+{
 	struct iphdr * iph = (struct iphdr *) packet;
 
 	iph -> ip_hlv = 0x45;
@@ -296,6 +287,7 @@ fill_iphdr(uint8_t * packet, uint16_t packetsize,
 /**
  * IPv4: Handles IPv4-packets according to Receive-handle diagram.
  *
+ * @param  fd         socket fd
  * @param  ip_packet  IP-packet to be handled
  * @param  packetsize Length of the packet
  * @return            ZERO - packet handled successfully;
@@ -303,8 +295,7 @@ fill_iphdr(uint8_t * packet, uint16_t packetsize,
  * @see               receive_ether
  * @see               iphdr
  */
-int8_t
-handle_ipv4(uint8_t * ip_packet, int32_t packetsize)
+int8_t handle_ipv4(int fd, uint8_t * ip_packet, uint32_t packetsize)
 {
 	struct iphdr * iph;
 	int32_t old_sum;
@@ -373,10 +364,10 @@ handle_ipv4(uint8_t * ip_packet, int32_t packetsize)
 
 	switch (iph -> ip_p) {
 	case IPTYPE_ICMP:
-		return handle_icmp(iph, ip_packet + sizeof(struct iphdr),
+		return handle_icmp(fd, iph, ip_packet + sizeof(struct iphdr),
 		                   iph -> ip_len - sizeof(struct iphdr));
 	case IPTYPE_UDP:
-		return handle_udp(ip_packet + sizeof(struct iphdr),
+		return handle_udp(fd, ip_packet + sizeof(struct iphdr),
 		                  iph -> ip_len - sizeof(struct iphdr));
 	case IPTYPE_TCP:
 		return handle_tcp(ip_packet + sizeof(struct iphdr),
@@ -398,13 +389,14 @@ handle_ipv4(uint8_t * ip_packet, int32_t packetsize)
  *       - IPv4 checksum is calculaded.
  *       - If payload type is UDP, then the UDP checksum is calculated also.
  *
- *       We sent an ARP request first, if this is the first packet sent to
+ *       We send an ARP request first, if this is the first packet sent to
  *       the declared IPv4 destination address. In this case we store the
- *       the packet and sent it later if we receive the ARP response.
- *       If the MAC address is known already, then we send the packet immediatly.
+ *       the packet and send it later if we receive the ARP response.
+ *       If the MAC address is known already, then we send the packet immediately.
  *       If there is already an ARP request pending, then we drop this packet
  *       and send again an ARP request.
  *
+ * @param  fd         socket fd
  * @param  ip_packet  IP-packet to be handled
  * @param  packetsize Length of the packet
  * @return            -2 - packet dropped (MAC address not resolved - ARP request pending)
@@ -416,12 +408,12 @@ handle_ipv4(uint8_t * ip_packet, int32_t packetsize)
  * @see               receive_ether
  * @see               iphdr
  */
-int
-send_ipv4(void* buffer, int len)
+int send_ipv4(int fd, void* buffer, int len)
 {
-	arp_entry_t *arp_entry;
+	arp_entry_t *arp_entry = 0;
 	struct iphdr *ip;
 	const uint8_t *mac_addr = 0;
+	uint32_t ip_dst = 0;
 
 	if(len + sizeof(struct ethhdr) > ETH_MTU_SIZE)
 		return -1;
@@ -453,6 +445,7 @@ send_ipv4(void* buffer, int len)
 		fill_udp_checksum(ip);
 	}
 
+	ip_dst = ip->ip_dst;
 	// Check if the MAC address is already cached
 	if(~ip->ip_dst == 0
 	|| ( ((~subnet_mask) & ip->ip_dst) == ~subnet_mask &&
@@ -469,8 +462,10 @@ send_ipv4(void* buffer, int len)
 		if((subnet_mask & own_ip) == (subnet_mask & ip->ip_dst))
 			arp_entry = lookup_mac_addr(ip->ip_dst);
 		// if not then we need to know the router's IP address
-		else
+		else {
+			ip_dst = router_ip;
 			arp_entry = lookup_mac_addr(router_ip);
+		}
 		if(arp_entry && memcmp(arp_entry->mac_addr, null_mac_addr, 6) != 0)
 			mac_addr = arp_entry->mac_addr;
 	}
@@ -478,7 +473,7 @@ send_ipv4(void* buffer, int len)
 	// If we could not resolv the MAC address by our own...
 	if(!mac_addr) {
 		// send the ARP request
-		arp_send_request(ip->ip_dst);
+		arp_send_request(fd, ip_dst);
 
 		// drop the current packet if there is already a ARP request pending
 		if(arp_entry)
@@ -493,13 +488,21 @@ send_ipv4(void* buffer, int len)
 			arp_consumer = (arp_consumer+1)%ARP_ENTRIES;
 
 		// store the packet to be send if the ARP reply is received
-		arp_entry->ipv4_addr = ip->ip_dst;
+		arp_entry->pkt_pending = 1;
+		arp_entry->ipv4_addr = ip_dst;
 		memset(arp_entry->mac_addr, 0, 6);
-		fill_ethhdr (arp_entry->eth_frame, htons(ETHERTYPE_IP),
+		fill_ethhdr (pending_pkt_frame, htons(ETHERTYPE_IP),
 		             get_mac_address(), null_mac_addr);
-		memcpy(&arp_entry->eth_frame[sizeof(struct ethhdr)],
+		memcpy(&pending_pkt_frame[sizeof(struct ethhdr)],
 		       buffer, len);
-		arp_entry->eth_len = len + sizeof(struct ethhdr);
+		pending_pkt_len = len + sizeof(struct ethhdr);
+
+		set_timer(TICKS_SEC);
+		do {
+			receive_ether(fd);
+			if (!arp_entry->eth_len)
+				break;
+		} while (get_timer() > 0);
 
 		return 0;
 	}
@@ -508,7 +511,7 @@ send_ipv4(void* buffer, int len)
 	fill_ethhdr(arp_entry->eth_frame, htons(ETHERTYPE_IP),
 	            get_mac_address(), mac_addr);
 	memcpy(&arp_entry->eth_frame[sizeof(struct ethhdr)], buffer, len);
-	return send_ether(arp_entry->eth_frame, len + sizeof(struct ethhdr));
+	return send_ether(fd, arp_entry->eth_frame, len + sizeof(struct ethhdr));
 }
 
 /**
@@ -518,11 +521,9 @@ send_ipv4(void* buffer, int len)
  *
  * @param  ipv4_hdr    Points to the place where IPv4-header starts.
  */
-
-static void
-fill_udp_checksum(struct iphdr *ipv4_hdr)
+static void fill_udp_checksum(struct iphdr *ipv4_hdr)
 {
-	int i;
+	unsigned i;
 	unsigned long checksum = 0;
 	struct iphdr ip_hdr;
 	char *ptr;
@@ -548,6 +549,13 @@ fill_udp_checksum(struct iphdr *ipv4_hdr)
 	checksum = (checksum >> 16) + (checksum & 0xffff);
 	checksum += (checksum >> 16);
 	udp_hdr->uh_sum = ~checksum;
+
+	/* As per RFC 768, if the computed  checksum  is zero,
+	 * it is transmitted as all ones (the equivalent in
+	 * one's complement arithmetic).
+	 */
+	if (udp_hdr->uh_sum == 0)
+		udp_hdr->uh_sum = ~udp_hdr->uh_sum;
 }
 
 /**
@@ -558,8 +566,7 @@ fill_udp_checksum(struct iphdr *ipv4_hdr)
  * @return            Checksum
  * @see               iphdr
  */
-static unsigned short
-checksum(unsigned short * packet, int words)
+static unsigned short checksum(unsigned short * packet, int words)
 {
 	unsigned long checksum;
 
@@ -571,8 +578,7 @@ checksum(unsigned short * packet, int words)
 	return ~checksum;
 }
 
-static arp_entry_t*
-lookup_mac_addr(uint32_t ipv4_addr)
+static arp_entry_t* lookup_mac_addr(uint32_t ipv4_addr)
 {
 	unsigned int i;
 
@@ -588,10 +594,10 @@ lookup_mac_addr(uint32_t ipv4_addr)
  * ARP: Sends an ARP-request package.
  *      For given IPv4 retrieves MAC via ARP (makes several attempts)
  *
+ * @param  fd        socket fd
  * @param  dest_ip   IP of the host which MAC should be obtained
  */
-static void
-arp_send_request(uint32_t dest_ip)
+static void arp_send_request(int fd, uint32_t dest_ip)
 {
 	arp_entry_t *arp_entry = &arp_table[arp_producer];
 
@@ -601,7 +607,7 @@ arp_send_request(uint32_t dest_ip)
 	fill_ethhdr(arp_entry->eth_frame, ETHERTYPE_ARP,
 	            get_mac_address(), broadcast_mac);
 
-	send_ether(arp_entry->eth_frame,
+	send_ether(fd, arp_entry->eth_frame,
 	     sizeof(struct ethhdr) + sizeof(struct arphdr));
 }
 
@@ -610,11 +616,11 @@ arp_send_request(uint32_t dest_ip)
  *      This package is used to serve foreign requests (in case IP in
  *      foreign request matches our host IP).
  *
+ * @param  fd        socket fd
  * @param  src_ip    requester IP address (foreign IP)
  * @param  src_mac   requester MAC address (foreign MAC)
  */
-static void
-arp_send_reply(uint32_t src_ip, uint8_t * src_mac)
+static void arp_send_reply(int fd, uint32_t src_ip, uint8_t * src_mac)
 {
 	arp_entry_t *arp_entry = &arp_table[arp_producer];
 
@@ -624,7 +630,7 @@ arp_send_reply(uint32_t src_ip, uint8_t * src_mac)
 	fill_arphdr(&arp_entry->eth_frame[sizeof(struct ethhdr)], ARP_REPLY,
 	            get_mac_address(), own_ip, src_mac, src_ip);
 
-	send_ether(arp_entry->eth_frame,
+	send_ether(fd, arp_entry->eth_frame,
 	     sizeof(struct ethhdr) + sizeof(struct arphdr));
 }
 
@@ -645,10 +651,9 @@ arp_send_reply(uint32_t src_ip, uint8_t * src_mac)
  * @see                arphdr
  * @see                fill_ethhdr
  */
-static void
-fill_arphdr(uint8_t * packet, uint8_t opcode,
-	    const uint8_t * src_mac, uint32_t src_ip,
-	    const uint8_t * dest_mac, uint32_t dest_ip)
+static void fill_arphdr(uint8_t * packet, uint8_t opcode,
+                        const uint8_t * src_mac, uint32_t src_ip,
+                        const uint8_t * dest_mac, uint32_t dest_ip)
 {
 	struct arphdr * arph = (struct arphdr *) packet;
 
@@ -668,6 +673,7 @@ fill_arphdr(uint8_t * packet, uint8_t opcode,
  * ARP: Handles ARP-messages according to Receive-handle diagram.
  *      Updates arp_table for outstanding ARP requests (see arp_getmac).
  *
+ * @param  fd         socket fd
  * @param  packet     ARP-packet to be handled
  * @param  packetsize length of the packet
  * @return            ZERO - packet handled successfully;
@@ -676,8 +682,7 @@ fill_arphdr(uint8_t * packet, uint8_t opcode,
  * @see               receive_ether
  * @see               arphdr
  */
-int8_t
-handle_arp(uint8_t * packet, int32_t packetsize)
+int8_t handle_arp(int fd, uint8_t * packet, uint32_t packetsize)
 {
 	struct arphdr * arph = (struct arphdr *) packet;
 
@@ -694,7 +699,7 @@ handle_arp(uint8_t * packet, int32_t packetsize)
 	case ARP_REQUEST:
 		// foreign request
 		if(own_ip != 0)
-			arp_send_reply(htonl(arph->src_ip), arph -> src_mac);
+			arp_send_reply(fd, htonl(arph->src_ip), arph -> src_mac);
 		return 0; // no error
 	case ARP_REPLY: {
 		unsigned int i;
@@ -723,11 +728,12 @@ handle_arp(uint8_t * packet, int32_t packetsize)
 		memcpy(arp_table[i].mac_addr, arph->src_mac, 6);
 
 		// do we have something to send
-		if(arp_table[i].eth_len > 0) {
-			struct ethhdr * ethh = (struct ethhdr *) arp_table[i].eth_frame;
+		if (arp_table[i].pkt_pending) {
+			struct ethhdr * ethh = (struct ethhdr *) pending_pkt_frame;
 			memcpy(ethh -> dest_mac, arp_table[i].mac_addr, 6);
 
-			send_ether(arp_table[i].eth_frame, arp_table[i].eth_len);
+			send_ether(fd, pending_pkt_frame, pending_pkt_len);
+			arp_table[i].pkt_pending = 0;
 			arp_table[i].eth_len = 0;
 		}
 		return 0; // no error
@@ -746,10 +752,10 @@ handle_arp(uint8_t * packet, int32_t packetsize)
  *       In other words, reading a value of 0 form this variable
  *       means that an answer to the request has been arrived.
  *
+ * @param  fd            socket descriptor
  * @param  _ping_dst_ip  destination IPv4 address
  */
-void
-ping_ipv4(uint32_t _ping_dst_ip)
+void ping_ipv4(int fd, uint32_t _ping_dst_ip)
 {
 	unsigned char packet[sizeof(struct iphdr) + sizeof(struct icmphdr)];
 	struct icmphdr *icmp;
@@ -772,7 +778,7 @@ ping_ipv4(uint32_t _ping_dst_ip)
 
 	icmp->checksum =
 	    checksum((unsigned short *) icmp, sizeof(struct icmphdr) >> 1);
-	send_ipv4(packet, sizeof(struct iphdr) + sizeof(struct icmphdr));
+	send_ipv4(fd, packet, sizeof(struct iphdr) + sizeof(struct icmphdr));
 }
 
 /**
@@ -782,8 +788,7 @@ ping_ipv4(uint32_t _ping_dst_ip)
  *
  * @return  ping_dst_ip  host IPv4 address
  */
-uint32_t
-pong_ipv4(void)
+uint32_t pong_ipv4(void)
 {
 	return ping_dst_ip;
 }
@@ -791,14 +796,15 @@ pong_ipv4(void)
 /**
  * ICMP: Handles ICMP-packets according to Receive-handle diagram.
  *
+ * @param  fd         socket fd
  * @param  icmp_packet  ICMP-packet to be handled
  * @param  packetsize   Length of the packet
  * @return              ZERO - packet handled successfully;
  *                      NON ZERO - packet was not handled (e.g. bad format)
  * @see                 handle_ipv4
  */
-static int8_t
-handle_icmp(struct iphdr * iph, uint8_t * packet, int32_t packetsize)
+static int8_t handle_icmp(int fd, struct iphdr * iph, uint8_t * packet,
+			  int32_t packetsize)
 {
 	struct icmphdr *icmp = (struct icmphdr *) packet;
 
@@ -851,7 +857,7 @@ handle_icmp(struct iphdr * iph, uint8_t * packet, int32_t packetsize)
 		reply_icmph->checksum = checksum((unsigned short *) reply_icmph,
 		                                 sizeof(struct icmphdr) >> 1);
 
-		send_ipv4(reply_packet, sizeof(struct iphdr) + packetsize);
+		send_ipv4(fd, reply_packet, sizeof(struct iphdr) + packetsize);
 		break;
 	}
 	case ICMP_TIME_EXCEEDED:
